package com.sun.org.apache.bcel.internal.classfile;

import java.util.Objects;
import java.util.Set;
import java.util.TreeSet;
import com.sun.org.apache.bcel.internal.util.ClassQueue;
import java.util.List;
import java.util.ArrayList;
import java.util.StringTokenizer;
import com.sun.org.apache.bcel.internal.generic.Type;
import java.io.OutputStream;
import java.io.ByteArrayOutputStream;
import java.io.DataOutputStream;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.File;
import com.sun.org.apache.bcel.internal.util.SyntheticRepository;
import com.sun.org.apache.bcel.internal.util.Repository;
import com.sun.org.apache.bcel.internal.util.BCELComparator;

public class JavaClass extends AccessFlags implements Cloneable, Node, Comparable<JavaClass>
{
    private String file_name;
    private String package_name;
    private String source_file_name;
    private int class_name_index;
    private int superclass_name_index;
    private String class_name;
    private String superclass_name;
    private int major;
    private int minor;
    private ConstantPool constant_pool;
    private int[] interfaces;
    private String[] interface_names;
    private Field[] fields;
    private Method[] methods;
    private Attribute[] attributes;
    private AnnotationEntry[] annotations;
    private byte source;
    private boolean isAnonymous;
    private boolean isNested;
    private boolean computedNestedTypeStatus;
    public static final byte HEAP = 1;
    public static final byte FILE = 2;
    public static final byte ZIP = 3;
    private static BCELComparator bcelComparator;
    private transient Repository repository;
    
    public JavaClass(int class_name_index, int superclass_name_index, String file_name, int major, int minor, int access_flags, ConstantPool constant_pool, int[] interfaces, Field[] fields, Method[] methods, Attribute[] attributes, byte source) {
        super(access_flags);
        this.source_file_name = "<Unknown>";
        this.source = 1;
        this.isAnonymous = false;
        this.isNested = false;
        this.computedNestedTypeStatus = false;
        this.repository = SyntheticRepository.getInstance();
        if (interfaces == null) {
            interfaces = new int[0];
        }
        if (attributes == null) {
            attributes = new Attribute[0];
        }
        if (fields == null) {
            fields = new Field[0];
        }
        if (methods == null) {
            methods = new Method[0];
        }
        this.class_name_index = class_name_index;
        this.superclass_name_index = superclass_name_index;
        this.file_name = file_name;
        this.major = major;
        this.minor = minor;
        this.constant_pool = constant_pool;
        this.interfaces = interfaces;
        this.fields = fields;
        this.methods = methods;
        this.attributes = attributes;
        this.source = source;
        int length = attributes.length;
        int j = 0;
        while (j < length) {
            Attribute attribute = attributes[j];
            if ((attribute instanceof SourceFile)) {
                this.source_file_name = ((SourceFile)attribute).getSourceFileName();
                break;
            }
            ++j;
        }
        this.class_name = constant_pool.getConstantString(class_name_index, (byte)7);
        this.class_name = Utility.compactClassName(this.class_name, false);
        int index = this.class_name.lastIndexOf(46);
        if (index < 0) {
            this.package_name = "";
        }
        else {
            this.package_name = this.class_name.substring(0, index);
        }
        if (superclass_name_index > 0) {
            this.superclass_name = constant_pool.getConstantString(superclass_name_index, (byte)7);
            this.superclass_name = Utility.compactClassName(this.superclass_name, false);
        }
        else {
            this.superclass_name = "java.lang.Object";
        }
        this.interface_names = new String[interfaces.length];
        for (int i = 0; i < interfaces.length; ++i) {
            String str = constant_pool.getConstantString(interfaces[i], (byte)7);
            this.interface_names[i] = Utility.compactClassName(str, false);
        }
    }
    
    public JavaClass(int class_name_index, int superclass_name_index, String file_name, int major, int minor, int access_flags, ConstantPool constant_pool, int[] interfaces, Field[] fields, Method[] methods, Attribute[] attributes) {
        this(class_name_index, superclass_name_index, file_name, major, minor, access_flags, constant_pool, interfaces, fields, methods, attributes, (byte)1);
    }
    
    @Override
    public void accept(Visitor v) {
        v.visitJavaClass(this);
    }
    
    public void dump(File file) throws IOException {
        String parent = file.getParent();
        if (parent != null) {
            File dir = new File(parent);
            if (!dir.mkdirs() && !dir.isDirectory()) {
                throw new IOException("Could not create the directory " + dir);
            }
        }
        try (final DataOutputStream dos = new DataOutputStream(new FileOutputStream(file))) {
            this.dump(dos);
        }
    }
    
    public void dump(String _file_name) throws IOException {
        this.dump(new File(_file_name));
    }
    
    public byte[] getBytes() {
        ByteArrayOutputStream s = new ByteArrayOutputStream();
        DataOutputStream ds = new DataOutputStream(s);
        try {
            this.dump(ds);
        }
        catch (IOException e) {
            System.err.println("Error dumping class: " + e.getMessage());
        }
        finally {
            try {
                ds.close();
            }
            catch (IOException e2) {
                System.err.println("Error dumping class: " + e2.getMessage());
            }
        }
        return s.toByteArray();
    }
    
    public void dump(OutputStream file) throws IOException {
        this.dump(new DataOutputStream(file));
    }
    
    public void dump(DataOutputStream file) throws IOException {
        file.writeInt(-889275714);
        file.writeShort(this.minor);
        file.writeShort(this.major);
        this.constant_pool.dump(file);
        file.writeShort(super.getAccessFlags());
        file.writeShort(this.class_name_index);
        file.writeShort(this.superclass_name_index);
        file.writeShort(this.interfaces.length);
        for (int interface1 : this.interfaces) {
            file.writeShort(interface1);
        }
        file.writeShort(this.fields.length);
        for (Field field : this.fields) {
            field.dump(file);
        }
        file.writeShort(this.methods.length);
        for (Method method : this.methods) {
            method.dump(file);
        }
        if (this.attributes != null) {
            file.writeShort(this.attributes.length);
            for (Attribute attribute : this.attributes) {
                attribute.dump(file);
            }
        }
        else {
            file.writeShort(0);
        }
        file.flush();
    }
    
    public Attribute[] getAttributes() {
        return this.attributes;
    }
    
    public AnnotationEntry[] getAnnotationEntries() {
        if (this.annotations == null) {
            this.annotations = AnnotationEntry.createAnnotationEntries(this.getAttributes());
        }
        return this.annotations;
    }
    
    public String getClassName() {
        return this.class_name;
    }
    
    public String getPackageName() {
        return this.package_name;
    }
    
    public int getClassNameIndex() {
        return this.class_name_index;
    }
    
    public ConstantPool getConstantPool() {
        return this.constant_pool;
    }
    
    public Field[] getFields() {
        return this.fields;
    }
    
    public String getFileName() {
        return this.file_name;
    }
    
    public String[] getInterfaceNames() {
        return this.interface_names;
    }
    
    public int[] getInterfaceIndices() {
        return this.interfaces;
    }
    
    public int getMajor() {
        return this.major;
    }
    
    public Method[] getMethods() {
        return this.methods;
    }
    
    public Method getMethod(java.lang.reflect.Method m) {
        Method[] methods = this.methods;
        int length = methods.length;
        int i = 0;
        while (i < length) {
            Method method = methods[i];
            if (m.getName().equals(method.getName()) && m.getModifiers() == method.getModifiers() && Type.getSignature(m).equals(method.getSignature())) {
                return method;
            }
            ++i;
        }
        return null;
    }
    
    public int getMinor() {
        return this.minor;
    }
    
    public String getSourceFileName() {
        return this.source_file_name;
    }
    
    public String getSuperclassName() {
        return this.superclass_name;
    }
    
    public int getSuperclassNameIndex() {
        return this.superclass_name_index;
    }
    
    public void setAttributes(Attribute[] attributes) {
        this.attributes = attributes;
    }
    
    public void setClassName(String class_name) {
        this.class_name = class_name;
    }
    
    public void setClassNameIndex(int class_name_index) {
        this.class_name_index = class_name_index;
    }
    
    public void setConstantPool(ConstantPool constant_pool) {
        this.constant_pool = constant_pool;
    }
    
    public void setFields(Field[] fields) {
        this.fields = fields;
    }
    
    public void setFileName(String file_name) {
        this.file_name = file_name;
    }
    
    public void setInterfaceNames(String[] interface_names) {
        this.interface_names = interface_names;
    }
    
    public void setInterfaces(int[] interfaces) {
        this.interfaces = interfaces;
    }
    
    public void setMajor(int major) {
        this.major = major;
    }
    
    public void setMethods(Method[] methods) {
        this.methods = methods;
    }
    
    public void setMinor(int minor) {
        this.minor = minor;
    }
    
    public void setSourceFileName(String source_file_name) {
        this.source_file_name = source_file_name;
    }
    
    public void setSuperclassName(String superclass_name) {
        this.superclass_name = superclass_name;
    }
    
    public void setSuperclassNameIndex(int superclass_name_index) {
        this.superclass_name_index = superclass_name_index;
    }
    
    @Override
    public String toString() {
        String access = Utility.accessToString(super.getAccessFlags(), true);
        access = (access.isEmpty() ? "" : (access + " "));
        StringBuilder buf = new StringBuilder(128);
        buf.append(access).append(Utility.classOrInterface(super.getAccessFlags())).append(" ").append(this.class_name).append(" extends ").append(Utility.compactClassName(this.superclass_name, false)).append('\n');
        int size = this.interfaces.length;
        if (size > 0) {
            buf.append("implements\t\t");
            for (int i = 0; i < size; ++i) {
                buf.append(this.interface_names[i]);
                if (i < size - 1) {
                    buf.append(", ");
                }
            }
            buf.append('\n');
        }
        buf.append("filename\t\t").append(this.file_name).append('\n');
        buf.append("compiled from\t\t").append(this.source_file_name).append('\n');
        buf.append("compiler version\t").append(this.major).append(".").append(this.minor).append('\n');
        buf.append("access flags\t\t").append(super.getAccessFlags()).append('\n');
        buf.append("constant pool\t\t").append(this.constant_pool.getLength()).append(" entries\n");
        buf.append("ACC_SUPER flag\t\t").append(this.isSuper()).append("\n");
        if (this.attributes.length > 0) {
            buf.append("\nAttribute(s):\n");
            for (Attribute attribute : this.attributes) {
                buf.append(indent(attribute));
            }
        }
        AnnotationEntry[] annotations = this.getAnnotationEntries();
        if (annotations != null && annotations.length > 0) {
            buf.append("\nAnnotation(s):\n");
            for (AnnotationEntry annotation : annotations) {
                buf.append(indent(annotation));
            }
        }
        if (this.fields.length > 0) {
            buf.append("\n").append(this.fields.length).append(" fields:\n");
            for (Field field : this.fields) {
                buf.append("\t").append(field).append('\n');
            }
        }
        if (this.methods.length > 0) {
            buf.append("\n").append(this.methods.length).append(" methods:\n");
            for (Method method : this.methods) {
                buf.append("\t").append(method).append('\n');
            }
        }
        return buf.toString();
    }
    
    private static String indent(Object obj) {
        StringTokenizer tok = new StringTokenizer(obj.toString(), "\n");
        StringBuilder buf = new StringBuilder();
        while (tok.hasMoreTokens()) {
            buf.append("\t").append(tok.nextToken()).append("\n");
        }
        return buf.toString();
    }
    
    public JavaClass copy() {
        JavaClass c = null;
        try {
            c = (JavaClass)this.clone();
            c.constant_pool = this.constant_pool.copy();
            c.interfaces = (int[])this.interfaces.clone();
            c.interface_names = (String[])this.interface_names.clone();
            c.fields = new Field[this.fields.length];
            for (int i = 0; i < this.fields.length; ++i) {
                c.fields[i] = this.fields[i].copy(c.constant_pool);
            }
            c.methods = new Method[this.methods.length];
            for (int i = 0; i < this.methods.length; ++i) {
                c.methods[i] = this.methods[i].copy(c.constant_pool);
            }
            c.attributes = new Attribute[this.attributes.length];
            for (int i = 0; i < this.attributes.length; ++i) {
                c.attributes[i] = this.attributes[i].copy(c.constant_pool);
            }
        }
        catch (CloneNotSupportedException ex) {}
        return c;
    }
    
    public final boolean isSuper() {
        return (super.getAccessFlags() & 0x20) != 0x0;
    }
    
    public final boolean isClass() {
        return (super.getAccessFlags() & 0x200) == 0x0;
    }
    
    public final boolean isAnonymous() {
        this.computeNestedTypeStatus();
        return this.isAnonymous;
    }
    
    public final boolean isNested() {
        this.computeNestedTypeStatus();
        return this.isNested;
    }
    
    private void computeNestedTypeStatus() {
        if (this.computedNestedTypeStatus) {
            return;
        }
        for (Attribute attribute : this.attributes) {
            if ((attribute instanceof InnerClasses)) {
                InnerClass[] innerClasses;
                for (InnerClass innerClasse : innerClasses = ((InnerClasses)attribute).getInnerClasses()) {
                    boolean innerClassAttributeRefersToMe = false;
                    String inner_class_name = this.constant_pool.getConstantString(innerClasse.getInnerClassIndex(), (byte)7);
                    inner_class_name = Utility.compactClassName(inner_class_name);
                    if (inner_class_name.equals(this.getClassName())) {
                        innerClassAttributeRefersToMe = true;
                    }
                    if (innerClassAttributeRefersToMe) {
                        this.isNested = true;
                        if (innerClasse.getInnerNameIndex() == 0) {
                            this.isAnonymous = true;
                        }
                    }
                }
            }
        }
        this.computedNestedTypeStatus = true;
        return;
    }
    
    public final byte getSource() {
        return this.source;
    }
    
    public Repository getRepository() {
        return this.repository;
    }
    
    public void setRepository(Repository repository) {
        this.repository = repository;
    }
    
    public final boolean instanceOf(JavaClass super_class) throws ClassNotFoundException {
        if (this.equals(super_class)) {
            return true;
        }
        JavaClass[] super_classes = this.getSuperClasses();
        int length = super_classes.length;
        int i = 0;
        while (i < length) {
            JavaClass super_classe = super_classes[i];
            if (super_classe.equals(super_class)) {
                return true;
            }
            ++i;
        }
        return super_class.isInterface() && this.implementationOf(super_class);
    }
    
    public boolean implementationOf(JavaClass inter) throws ClassNotFoundException {
        if (!inter.isInterface()) {
            throw new IllegalArgumentException(inter.getClassName() + " is no interface");
        }
        if (this.equals(inter)) {
            return true;
        }
        JavaClass[] super_interfaces = this.getAllInterfaces();
        int length = super_interfaces.length;
        int i = 0;
        while (i < length) {
            JavaClass super_interface = super_interfaces[i];
            if (super_interface.equals(inter)) {
                return true;
            }
            ++i;
        }
        return false;
    }
    
    public JavaClass getSuperClass() throws ClassNotFoundException {
        if ("java.lang.Object".equals(this.getClassName())) {
            return null;
        }
        return this.repository.loadClass(this.getSuperclassName());
    }
    
    public JavaClass[] getSuperClasses() throws ClassNotFoundException {
        JavaClass clazz = this;
        List<JavaClass> allSuperClasses = new ArrayList();
        for (clazz = clazz.getSuperClass(); clazz != null; clazz = clazz.getSuperClass()) {
            allSuperClasses.add(clazz);
        }
        return (JavaClass[])allSuperClasses.toArray(new JavaClass[allSuperClasses.size()]);
    }
    
    public JavaClass[] getInterfaces() throws ClassNotFoundException {
        String[] _interfaces = this.getInterfaceNames();
        JavaClass[] classes = new JavaClass[_interfaces.length];
        for (int i = 0; i < _interfaces.length; ++i) {
            classes[i] = this.repository.loadClass(_interfaces[i]);
        }
        return classes;
    }
    
    public JavaClass[] getAllInterfaces() throws ClassNotFoundException {
        ClassQueue queue = new ClassQueue();
        Set<JavaClass> allInterfaces = new TreeSet();
        queue.enqueue(this);
        while (!queue.empty()) {
            JavaClass clazz = queue.dequeue();
            JavaClass souper = clazz.getSuperClass();
            JavaClass[] _interfaces = clazz.getInterfaces();
            if (clazz.isInterface()) {
                allInterfaces.add(clazz);
            }
            else if (souper != null) {
                queue.enqueue(souper);
            }
            for (JavaClass _interface : _interfaces) {
                queue.enqueue(_interface);
            }
        }
        return (JavaClass[])allInterfaces.toArray(new JavaClass[allInterfaces.size()]);
    }
    
    public static BCELComparator getComparator() {
        return JavaClass.bcelComparator;
    }
    
    public static void setComparator(BCELComparator comparator) {
        JavaClass.bcelComparator = comparator;
    }
    
    @Override
    public boolean equals(Object obj) {
        return JavaClass.bcelComparator.equals(this, obj);
    }
    
    @Override
    public int compareTo(JavaClass obj) {
        return this.getClassName().compareTo(obj.getClassName());
    }
    
    @Override
    public int hashCode() {
        return JavaClass.bcelComparator.hashCode(this);
    }
    
    static {
        JavaClass.bcelComparator = new BCELComparator() {
            @Override
            public boolean equals(Object o1, Object o2) {
                JavaClass THIS = (JavaClass)o1;
                JavaClass THAT = (JavaClass)o2;
                return Objects.equals(THIS.getClassName(), THAT.getClassName());
            }
            
            @Override
            public int hashCode(Object o) {
                JavaClass THIS = (JavaClass)o;
                return THIS.getClassName().hashCode();
            }
        };
    }
}
